from datetime import datetime, timedelta, timezone
from unittest.mock import call, patch

from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings, tag
from django.urls import reverse
from regal import FileId, RegistrarCertification, RegistrarCertificationId
from regal.exceptions import InvalidData, RegistrarDoesNotExist

from ferda.constants import LogEntryType, LoggerResult, LogReferenceType
from ferda.registry.registrar.certification import AddCertificationParams, UpdateCertificationParams
from ferda.tests.utils import LoggerMixin

User = get_user_model()


def _make_certification(id: int, valid_days: int = 0) -> RegistrarCertification:
    # Python API model
    valid_from = datetime(2023, 1, 1, tzinfo=timezone.utc) + timedelta(days=id)
    return RegistrarCertification(
        certification_id=RegistrarCertificationId("Cert-" + str(id)),
        registrar_handle="Reg-1",
        valid_from=valid_from,
        valid_to=None if valid_days <= 0 else valid_from + timedelta(days=valid_days),
        classification=50,
        file_id=FileId("File " + str(id)),
    )


@tag("api", "registry")
@patch("ferda.registry.registrar.certification.REGISTRAR", autospec=True)
@override_settings(ROOT_URLCONF="ferda.tests.urls")
class GetCertificationTest(LoggerMixin, TestCase):
    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    def test_get_certifications(self, client):
        certifications = [_make_certification(i + 1) for i in range(3)]
        client.get_registrar_certifications.return_value = certifications

        response = self.client.get(reverse("api-1:registrar_certification", kwargs={"registrar_id": "Reg-1"}))

        self.assertEqual(response.status_code, 200)
        self.assertEqual([RegistrarCertification(**i) for i in response.json()], certifications)
        self.assertEqual(client.mock_calls, [call.get_registrar_certifications("Reg-1")])
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_VIEW,
            LoggerResult.SUCCESS,
            input_properties={"registrar_id": "Reg-1"},
            output_properties={"items_count": str(len(certifications))},
            output_references={LogReferenceType.REGISTRAR: "Reg-1"},
        )

    def test_get_certifications_empty(self, client):
        client.get_registrar_certifications.return_value = []

        response = self.client.get(reverse("api-1:registrar_certification", kwargs={"registrar_id": "Reg-1"}))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), [])
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_VIEW,
            LoggerResult.SUCCESS,
            input_properties={"registrar_id": "Reg-1"},
            output_references={LogReferenceType.REGISTRAR: "Reg-1"},
        )

    def test_get_certifications_not_found(self, client):
        client.get_registrar_certifications.side_effect = RegistrarDoesNotExist

        response = self.client.get(reverse("api-1:registrar_certification", kwargs={"registrar_id": "Reg-1"}))

        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Registrar does not exist"})
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_VIEW,
            LoggerResult.FAIL,
            input_properties={"registrar_id": "Reg-1"},
        )


@tag("api", "registry")
@patch("ferda.registry.registrar.certification.REGISTRAR_ADMIN", autospec=True)
@override_settings(ROOT_URLCONF="ferda.tests.urls")
class AddCertificationTest(LoggerMixin, TestCase):
    """Test class for registrar certifications."""

    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    # --- add_registrar_certification

    def test_add_certification_ok(self, client):
        params = AddCertificationParams(
            classification=50,
            file_id="File-1",
            valid_from=datetime(2023, 1, 1, 12, 0, 0, 0, tzinfo=timezone.utc),
            valid_to=datetime(2023, 6, 30, 0, 0, 0, tzinfo=timezone.utc)
        )
        client.add_registrar_certification.return_value = None

        response = self.client.post(
            reverse("api-1:registrar_certification", kwargs={"registrar_id": "Reg-1"}),
            data=params.dict(),
            content_type="application/json",
        )

        self.assertEqual(client.mock_calls, [call.add_registrar_certification("Reg-1", **params.dict())])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"")
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_CREATE,
            LoggerResult.SUCCESS,
            input_properties={
                "registrar_id": "Reg-1",
                "classification": str(params.classification),
                "file_id": params.file_id,
                "valid_from": params.valid_from.isoformat(),  # type: ignore [union-attr]
                "valid_to": params.valid_to.isoformat(),  # type: ignore [union-attr]
            },
            output_references={LogReferenceType.REGISTRAR: "Reg-1"},
        )

    def test_add_certification_not_found(self, client):
        params = AddCertificationParams(classification=50, file_id="File-1")
        client.add_registrar_certification.side_effect = RegistrarDoesNotExist

        response = self.client.post(
            reverse("api-1:registrar_certification", kwargs={"registrar_id": "Reg-1"}),
            data=params.dict(),
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Registrar does not exist"})
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_CREATE,
            LoggerResult.FAIL,
            input_properties={
                "registrar_id": "Reg-1",
                "classification": str(params.classification),
                "file_id": params.file_id,
            },
        )

    def test_add_certification_invalid(self, client):
        params = AddCertificationParams(classification=-50, file_id="File-1")
        client.add_registrar_certification.side_effect = InvalidData(fields=["classification"])

        response = self.client.post(
            reverse("api-1:registrar_certification", kwargs={"registrar_id": "Reg-1"}),
            data=params.dict(),
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Invalid data", "fields": ["classification"]})
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_CREATE,
            LoggerResult.FAIL,
            input_properties={
                "registrar_id": "Reg-1",
                "classification": str(params.classification),
                "file_id": params.file_id,
            },
        )


@tag("api", "registry")
@patch("ferda.registry.registrar.certification.REGISTRAR", autospec=True)
@patch("ferda.registry.registrar.certification.REGISTRAR_ADMIN", autospec=True)
@override_settings(ROOT_URLCONF="ferda.tests.urls")
class UpdateDeleteCertificationTest(LoggerMixin, TestCase):
    """Test class for registrar certifications."""

    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    # --- update_registrar_certification

    def test_update_certification_ok(self, admin_client, client):
        params = UpdateCertificationParams(
            classification=50,
            file_id="File-1",
            valid_to=datetime(2023, 6, 30, 0, 0, 0, tzinfo=timezone.utc)
        )
        client.get_registrar_certifications.return_value = [_make_certification(1)]
        admin_client.update_registrar_certification.return_value = None

        response = self.client.put(
            reverse("api-1:registrar_certification", kwargs={"registrar_id": "Reg-1", "certification_id": "Cert-1"}),
            data=params.dict(),
            content_type="application/json",
        )

        self.assertEqual(client.mock_calls, [call.get_registrar_certifications("Reg-1")])
        self.assertEqual(admin_client.mock_calls, [call.update_registrar_certification("Cert-1", **params.dict())])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"")
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_UPDATE,
            LoggerResult.SUCCESS,
            input_properties={
                "registrar_id": "Reg-1",
                "certification_id": "Cert-1",
                "set_classification": str(params.classification),
                "set_file_id": params.file_id,
                "set_valid_to": params.valid_to.isoformat(),  # type: ignore [union-attr]
            },
            output_references={LogReferenceType.REGISTRAR: "Reg-1"},
        )

    def test_update_certification_cert_not_found(self, admin_client, client):
        params = UpdateCertificationParams()
        client.get_registrar_certifications.return_value = []

        response = self.client.put(
            reverse("api-1:registrar_certification", kwargs={"registrar_id": "Reg-1", "certification_id": "Cert-1"}),
            data=params.dict(),
            content_type="application/json",
        )

        self.assertEqual(admin_client.mock_calls, [])
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Certification does not exist"})
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_UPDATE,
            LoggerResult.FAIL,
            input_properties={"registrar_id": "Reg-1", "certification_id": "Cert-1"},
        )

    def test_update_certification_reg_not_found(self, admin_client, client):
        params = UpdateCertificationParams()
        client.get_registrar_certifications.side_effect = RegistrarDoesNotExist

        response = self.client.put(
            reverse("api-1:registrar_certification", kwargs={"registrar_id": "Reg-1", "certification_id": "Cert-1"}),
            data=params.dict(),
            content_type="application/json",
        )

        self.assertEqual(admin_client.mock_calls, [])
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Registrar does not exist"})
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_UPDATE,
            LoggerResult.FAIL,
            input_properties={"registrar_id": "Reg-1", "certification_id": "Cert-1"},
        )

    def test_update_certification_invalid(self, admin_client, client):
        params = UpdateCertificationParams(classification=-50)
        client.get_registrar_certifications.return_value = [_make_certification(1)]
        admin_client.update_registrar_certification.side_effect = InvalidData(fields=["classification"])

        response = self.client.put(
            reverse("api-1:registrar_certification", kwargs={"registrar_id": "Reg-1", "certification_id": "Cert-1"}),
            data=params.dict(),
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Invalid data", "fields": ["classification"]})
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_UPDATE,
            LoggerResult.FAIL,
            input_properties={
                "registrar_id": "Reg-1",
                "certification_id": "Cert-1",
                "set_classification": str(params.classification),
            },
        )

    # --- delete_registrar_certification

    def test_delete_certification_ok(self, admin_client, client):
        client.get_registrar_certifications.return_value = [_make_certification(1)]
        admin_client.delete_registrar_certification.return_value = None

        response = self.client.delete(reverse("api-1:registrar_certification", kwargs={
            "registrar_id": "Reg-1", "certification_id": "Cert-1"
        }))

        self.assertEqual(client.mock_calls, [call.get_registrar_certifications("Reg-1")])
        self.assertEqual(admin_client.mock_calls, [call.delete_registrar_certification("Cert-1")])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"")
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_DELETE,
            LoggerResult.SUCCESS,
            input_properties={"registrar_id": "Reg-1", "certification_id": "Cert-1"},
            output_references={LogReferenceType.REGISTRAR: "Reg-1"},
        )

    def test_delete_certification_cert_not_found(self, admin_client, client):
        client.get_registrar_certifications.return_value = []

        response = self.client.delete(reverse("api-1:registrar_certification", kwargs={
            "registrar_id": "Reg-1", "certification_id": "Cert-1"
        }))

        self.assertEqual(admin_client.mock_calls, [])
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Certification does not exist"})
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_DELETE,
            LoggerResult.FAIL,
            input_properties={"registrar_id": "Reg-1", "certification_id": "Cert-1"},
        )

    def test_delete_certification_reg_not_found(self, admin_client, client):
        client.get_registrar_certifications.side_effect = RegistrarDoesNotExist

        response = self.client.delete(reverse("api-1:registrar_certification", kwargs={
            "registrar_id": "Reg-1", "certification_id": "Cert-1"
        }))

        self.assertEqual(admin_client.mock_calls, [])
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Registrar does not exist"})
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_DELETE,
            LoggerResult.FAIL,
            input_properties={"registrar_id": "Reg-1", "certification_id": "Cert-1"},
        )

    def test_delete_certification_invalid(self, admin_client, client):
        client.get_registrar_certifications.return_value = [_make_certification(1)]
        admin_client.delete_registrar_certification.side_effect = InvalidData(fields=["certification_id"])

        response = self.client.delete(reverse("api-1:registrar_certification", kwargs={
            "registrar_id": "Reg-1", "certification_id": "Cert-1"
        }))

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Invalid data", "fields": ["certification_id"]})
        self.assertLog(
            LogEntryType.REGISTRAR_CERTIFICATION_DELETE,
            LoggerResult.FAIL,
            input_properties={"registrar_id": "Reg-1", "certification_id": "Cert-1"},
        )
