from datetime import datetime
from unittest.mock import patch
from uuid import UUID

import pytz
from django.test import TestCase, override_settings
from django.urls import reverse
from grill import LogProperty

from ferda.constants import LogEntryType, LoggerResult
from ferda.models import User
from ferda.tests.utils import LoggerMixin
from ferda.views.ajax.search import RelatedContactsMixin, SearchContactView, SearchDomainView, SearchNssetView


@override_settings(ROOT_URLCONF='ferda.tests.urls')
class TestSearchContactView(LoggerMixin, TestCase):
    """Test SearchContactView."""

    def setUp(self):
        super().setUp()
        self.user = User.objects.create_superuser('admin', 'admin@example.com', 'password')
        self.client.force_login(self.user)

    @patch('ferda.backend.BACKEND.get_contact_info')
    @patch('ferda.backend.BACKEND.search_contact')
    def test_get(self, search_mock, info_mock):
        search_mock.return_value = {
            'results': [{
                'object_id': UUID(int=42),
                'matched_items': ['name', 'place.city'],
            }],
            'searched_items': ['name', 'email'],
        }
        info_mock.return_value = {
            'contact_handle': 'HARRY',
            'name': 'Harry Potter',
            'email': 'harry.potter@example.com',
        }
        response = self.client.get(
            reverse('registry:search:ajax_contact'),
            data={'query': 'harry', 'type': 'contact', 'items': ['name', 'email', 'unknown']},
        )
        data = response.json()
        self.assertEqual(
            data, {
                'search_results': [{
                    'object_id': str(UUID(int=42)),
                    'contact_handle': 'HARRY',
                    'name': 'Harry Potter',
                    'organization': None,
                    'telephone': None,
                    'email': 'harry.potter@example.com',
                    'place': None,
                    'matched_items': ['name', 'place.city'],
                    'match_keys': {'name': ['name'], 'place': ['place.city']},
                }]
            }
        )
        # Check logs
        search_items = [LogProperty('name', None), LogProperty('email', None), LogProperty('notify_email', None)]
        properties = {'query_values': [LogProperty('harry', None)], 'search_item': search_items}
        self.assertLog(
            LogEntryType.CONTACT_SEARCH,
            LoggerResult.SUCCESS,
            input_properties=properties,
            references={'contact': [str(UUID(int=42))]},
        )

    def test_get_empty(self):
        response = self.client.get(reverse('registry:search:ajax_contact'))
        self.assertJSONEqual(response.content, {'search_results': []})
        self.assertNoLogs()

    @override_settings(USE_TZ=True)
    @patch('ferda.backend.BACKEND.get_contact_info', autospec=True)
    @patch('ferda.backend.BACKEND.get_contact_history', autospec=True)
    @patch('ferda.backend.BACKEND.search_contact_history', autospec=True)
    def test_get_history(self, search_mock, history_mock, info_mock):
        search_mock.return_value = {
            'results': [{
               'object_id': UUID(int=42),
               'histories': [{
                   'matched_items': ['email'],
                   'valid_from': datetime(2019, 1, 1, tzinfo=pytz.utc),
                   'valid_to': None,
                   'contact_history_ids': [UUID(int=142)],
               }],
            }],
        }
        history_mock.return_value = {
            'timeline': [{
                'contact_history_id': UUID(int=142),
                'valid_from': datetime(2018, 1, 1, tzinfo=pytz.utc),
            }, {
                'contact_history_id': UUID(int=42),
                'valid_from': datetime(2019, 1, 1, tzinfo=pytz.utc),
            }],
            'valid_to': None,
        }
        info_mock.return_value = {
            'contact_handle': 'IDAHO',
            'name': 'Duncan Idaho',
            'email': 'duncan.idaho@example.com',
            'events': {
                'registered': {
                    'registrar_handle': 'REG-DUNE',
                    'timestamp': datetime(2018, 5, 5, tzinfo=pytz.utc),
                },
                'unregistered': None,
            },
        }
        response = self.client.get(
            reverse('registry:search:ajax_contact_history'),
            data={
                'query': 'idaho',
                'type': 'contact',
                'valid_from': '2019-01-01T00:00:00Z',
                'valid_to': '2020-01-01T00:00:00Z',
            },
        )
        data = response.json()
        self.assertEqual(data, {
            'search_results': [{
                'object_id': str(UUID(int=42)),
                'contact_handle': 'IDAHO',
                'name': 'Duncan Idaho',
                'organization': None,
                'email': 'duncan.idaho@example.com',
                'telephone': None,
                'events': {
                    'registered': {
                        'registrar_handle': 'REG-DUNE',
                        'timestamp': '2018-05-05T00:00:00Z',
                    },
                    'unregistered': None,
                },
                'histories': [{
                    'contact_handle': 'IDAHO',
                    'contact_history_ids': [str(UUID(int=142))],
                    'match_keys': {'email': ['email']},
                    'matched_items': ['email'],
                    'valid_from': '2019-01-01T00:00:00Z',
                    'valid_to': None,
                    'name': 'Duncan Idaho',
                    'organization': None,
                    'email': 'duncan.idaho@example.com',
                    'telephone': None,
                }],
            }],
        })

        # Check logs
        properties = {'query_values': [LogProperty('idaho', None)], 'search_item': []}
        self.assertLog(
            LogEntryType.CONTACT_SEARCH_HISTORY,
            LoggerResult.SUCCESS,
            input_properties=properties,
            references={'contact': [str(UUID(int=42))]},
        )

    def test_get_match_keys(self):
        get_match_keys = SearchContactView.get_match_keys
        self.assertEqual(
            get_match_keys([]),
            {},
        )
        self.assertEqual(
            get_match_keys(['name']),
            {'name': ['name']},
        )
        self.assertEqual(
            get_match_keys(['place.street', 'place.city']),
            {'place': ['place.street', 'place.city']},
        )


@override_settings(ROOT_URLCONF='ferda.tests.urls')
class TestSearchDomainView(LoggerMixin, TestCase):
    """Test SearchDomainView."""

    def setUp(self):
        super().setUp()
        self.user = User.objects.create_superuser('admin', 'admin@example.com', 'password')
        self.client.force_login(self.user)

    @patch('ferda.backend.BACKEND.get_contact_info')
    @patch('ferda.backend.BACKEND.get_domain_info')
    @patch('ferda.backend.BACKEND.search_domain')
    def test_get(self, search_mock, domain_mock, contact_mock):
        search_mock.return_value = {
            'results': [{
                'object_id': UUID(int=42),
                'matched_items': ['fqdn'],
            }],
            'searched_items': ['fqdn'],
        }
        domain_mock.return_value = {
            'fqdn': 'example.com',
            'expires_at': datetime(2019, 1, 1, 10, 0, 0, tzinfo=pytz.utc),
            'registrant': UUID(int=45),
        }
        contact_mock.return_value = {
            'contact_id': UUID(int=45),
            'contact_handle': 'HARRY',
            'name': 'Harry Potter',
        }
        response = self.client.get(
            reverse('registry:search:ajax_domain'),
            data={'query': 'example.com', 'type': 'domain', 'items': ['fqdn']},
        )
        data = response.json()
        self.assertEqual(
            data, {
                'search_results': [{
                    'object_id': str(UUID(int=42)),
                    'fqdn': 'example.com',
                    'expires_at': '2019-01-01T10:00:00Z',
                    'registrant': {
                        'name': 'Harry Potter',
                        'contact_handle': 'HARRY',
                        'contact_id': str(UUID(int=45)),
                    },
                    'matched_items': ['fqdn'],
                    'match_keys': {'fqdn': ['fqdn']},
                }]
            }
        )
        # Check logs
        search_items = [LogProperty('fqdn', None)]
        properties = {'query_values': [LogProperty('example.com', None)], 'search_item': search_items}
        self.assertLog(
            LogEntryType.DOMAIN_SEARCH,
            LoggerResult.SUCCESS,
            input_properties=properties,
            references={'domain': [str(UUID(int=42))]},
        )

    @patch('ferda.backend.BACKEND.get_domain_info')
    @patch('ferda.backend.BACKEND.search_domain')
    def test_no_registrant(self, search_mock, domain_mock):
        search_mock.return_value = {
            'results': [{
                'object_id': UUID(int=42),
                'matched_items': ['fqdn'],
            }],
            'searched_items': ['fqdn'],
        }
        domain_mock.return_value = {
            'fqdn': 'example.com',
            'expires_at': datetime(2019, 1, 1, 10, 0, 0, tzinfo=pytz.utc),
        }
        response = self.client.get(
            reverse('registry:search:ajax_domain'),
            data={'query': 'example.com', 'type': 'domain', 'items': ['fqdn']},
        )
        data = response.json()
        self.assertEqual(
            data, {
                'search_results': [{
                    'object_id': str(UUID(int=42)),
                    'fqdn': 'example.com',
                    'expires_at': '2019-01-01T10:00:00Z',
                    'matched_items': ['fqdn'],
                    'match_keys': {'fqdn': ['fqdn']},
                }]
            }
        )
        # Check logs
        search_items = [LogProperty('fqdn', None)]
        properties = {'query_values': [LogProperty('example.com', None)], 'search_item': search_items}
        self.assertLog(
            LogEntryType.DOMAIN_SEARCH,
            LoggerResult.SUCCESS,
            input_properties=properties,
            references={'domain': [str(UUID(int=42))]},
        )

    def test_get_empty(self):
        response = self.client.get(reverse('registry:search:ajax_domain'))
        self.assertJSONEqual(response.content, {'search_results': []})
        self.assertNoLogs()

    @override_settings(USE_TZ=True)
    @patch('ferda.backend.BACKEND.get_domain_info', autospec=True)
    @patch('ferda.backend.BACKEND.get_domain_history', autospec=True)
    @patch('ferda.backend.BACKEND.search_domain_history', autospec=True)
    def test_get_history(self, search_mock, history_mock, info_mock):
        search_mock.return_value = {
            'results': [{
               'object_id': UUID(int=42),
               'histories': [{
                   'matched_items': ['fqdn'],
                   'valid_from': datetime(2019, 1, 1, tzinfo=pytz.utc),
                   'valid_to': datetime(2020, 1, 1, tzinfo=pytz.utc),
                   'domain_history_ids': [UUID(int=142)],
               }],
            }],
        }
        history_mock.return_value = {
            'timeline': [{
                'domain_history_id': UUID(int=142),
                'valid_from': datetime(2018, 1, 1, tzinfo=pytz.utc),
            }, {
                'domain_history_id': UUID(int=42),
                'valid_from': datetime(2019, 1, 1, tzinfo=pytz.utc),
            }],
            'valid_to': None,
        }
        info_mock.return_value = {
            'fqdn': 'idaho.cz',
            'expires_at': datetime(2020, 1, 1, tzinfo=pytz.utc),
            'events': {
                'registered': {
                    'registrar_handle': 'REG-DUNE',
                    'timestamp': datetime(2018, 5, 5, tzinfo=pytz.utc),
                },
                'unregistered': None,
            },
        }
        response = self.client.get(
            reverse('registry:search:ajax_domain_history'),
            data={
                'query': 'idaho',
                'type': 'domain',
                'valid_from': '2019-01-01T00:00:00Z',
                'valid_to': '2020-01-01T00:00:00Z',
            },
        )
        data = response.json()
        self.assertEqual(data, {
            'search_results': [{
                'object_id': str(UUID(int=42)),
                'fqdn': 'idaho.cz',
                'expires_at': '2020-01-01T00:00:00Z',
                'events': {
                    'registered': {
                        'registrar_handle': 'REG-DUNE',
                        'timestamp': '2018-05-05T00:00:00Z',
                    },
                    'unregistered': None,
                },
                'histories': [{
                    'fqdn': 'idaho.cz',
                    'domain_history_ids': [str(UUID(int=142))],
                    'match_keys': {'fqdn': ['fqdn']},
                    'matched_items': ['fqdn'],
                    'valid_from': '2019-01-01T00:00:00Z',
                    'valid_to': '2020-01-01T00:00:00Z',
                    'expires_at': '2020-01-01T00:00:00Z',
                }],
            }],
        })

        # Check logs
        properties = {'query_values': [LogProperty('idaho', None)], 'search_item': [LogProperty('fqdn', None)]}
        self.assertLog(
            LogEntryType.DOMAIN_SEARCH_HISTORY,
            LoggerResult.SUCCESS,
            input_properties=properties,
            references={'domain': [str(UUID(int=42))]},
        )

    def test_get_match_keys(self):
        get_match_keys = SearchDomainView.get_match_keys
        self.assertEqual(
            get_match_keys([]),
            {},
        )
        self.assertEqual(
            get_match_keys(['fqdn']),
            {'fqdn': ['fqdn']},
        )


@override_settings(ROOT_URLCONF='ferda.tests.urls')
class TestSearchNssetView(LoggerMixin, TestCase):
    """Test SearchNssetView."""

    def setUp(self):
        super().setUp()
        self.user = User.objects.create_superuser('admin', 'admin@example.com', 'password')
        self.client.force_login(self.user)

    @patch('ferda.backend.BACKEND.get_nsset_info')
    @patch('ferda.backend.BACKEND.search_nsset')
    def test_get(self, search_mock, nsset_mock):
        search_mock.return_value = {
            'results': [{
                'object_id': UUID(int=42),
                'matched_items': ['dns_hosts.ip_addresses'],
            }],
            'searched_items': ['dns_hosts.ip_addresses'],
        }
        nsset_mock.return_value = {
            'nsset_handle': 'NSSET',
            'dns_hosts': {
                'fqdn': 'localhost',
                'ip_addresses': ['127.0.0.1'],
            }
        }
        response = self.client.get(
            reverse('registry:search:ajax_nsset'),
            data={'query': ['127.0.0.1'], 'type': 'nsset', 'items': ['dns_hosts.ip_addresses']},
        )
        data = response.json()
        self.assertEqual(
            data, {
                'search_results': [{
                    'object_id': str(UUID(int=42)),
                    'nsset_handle': 'NSSET',
                    'dns_hosts': {
                        'fqdn': 'localhost',
                        'ip_addresses': ['127.0.0.1'],
                    },
                    'technical_contacts': [],
                    'matched_items': ['dns_hosts.ip_addresses'],
                    'match_keys': {'dns_hosts': ['dns_hosts.ip_addresses']},
                }]
            }
        )
        # Check logs
        search_items = [LogProperty('dns_hosts.ip_addresses', None)]
        properties = {'query_values': [LogProperty('127.0.0.1', None)], 'search_item': search_items}
        self.assertLog(
            LogEntryType.NSSET_SEARCH,
            LoggerResult.SUCCESS,
            input_properties=properties,
            references={'nsset': [str(UUID(int=42))]},
        )

    def test_get_empty(self):
        response = self.client.get(reverse('registry:search:ajax_nsset'))
        self.assertJSONEqual(response.content, {'search_results': []})
        self.assertNoLogs()

    @override_settings(USE_TZ=True)
    @patch('ferda.backend.BACKEND.get_nsset_info', autospec=True)
    @patch('ferda.backend.BACKEND.get_nsset_history', autospec=True)
    @patch('ferda.backend.BACKEND.search_nsset_history', autospec=True)
    def test_get_history(self, search_mock, history_mock, info_mock):
        search_mock.return_value = {
            'results': [{
               'object_id': UUID(int=42),
               'histories': [{
                   'matched_items': ['nsset_handle'],
                   'valid_from': datetime(2019, 1, 1, tzinfo=pytz.utc),
                   'valid_to': datetime(2020, 1, 1, tzinfo=pytz.utc),
                   'nsset_history_ids': [UUID(int=142)],
               }],
            }],
        }
        history_mock.return_value = {
            'timeline': [{
                'nsset_history_id': UUID(int=142),
                'valid_from': datetime(2018, 1, 1, tzinfo=pytz.utc),
            }, {
                'nsset_history_id': UUID(int=42),
                'valid_from': datetime(2019, 1, 1, tzinfo=pytz.utc),
            }],
            'valid_to': None,
        }
        info_mock.return_value = {
            'nsset_handle': 'DUNE-NSSET',
            'dns_hosts': {
                'fqdn': 'localhost',
                'ip_addresses': ['127.0.0.1'],
            },
            'technical_contacts': [],
            'events': {
                'registered': {
                    'registrar_handle': 'REG-DUNE',
                    'timestamp': datetime(2018, 5, 5, tzinfo=pytz.utc),
                },
                'unregistered': None,
            },
        }
        response = self.client.get(
            reverse('registry:search:ajax_nsset_history'),
            data={
                'query': 'dune',
                'type': 'nsset',
                'valid_from': '2019-01-01T00:00:00Z',
                'valid_to': '2020-01-01T00:00:00Z',
            },
        )
        data = response.json()
        self.assertEqual(data, {
            'search_results': [{
                'object_id': str(UUID(int=42)),
                'nsset_handle': 'DUNE-NSSET',
                'technical_contacts': [],
                'dns_hosts': {
                    'fqdn': 'localhost',
                    'ip_addresses': ['127.0.0.1'],
                },
                'events': {
                    'registered': {
                        'registrar_handle': 'REG-DUNE',
                        'timestamp': '2018-05-05T00:00:00Z',
                    },
                    'unregistered': None,
                },
                'histories': [{
                    'nsset_handle': 'DUNE-NSSET',
                    'nsset_history_ids': [str(UUID(int=142))],
                    'match_keys': {'nsset_handle': ['nsset_handle']},
                    'matched_items': ['nsset_handle'],
                    'valid_from': '2019-01-01T00:00:00Z',
                    'valid_to': '2020-01-01T00:00:00Z',
                    'technical_contacts': [],
                    'dns_hosts': {
                        'fqdn': 'localhost',
                        'ip_addresses': ['127.0.0.1'],
                    },
                }],
            }],
        })

        # Check logs
        properties = {'query_values': [LogProperty('dune', None)], 'search_item': []}
        self.assertLog(
            LogEntryType.NSSET_SEARCH_HISTORY,
            LoggerResult.SUCCESS,
            input_properties=properties,
            references={'nsset': [str(UUID(int=42))]},
        )

    def test_get_match_keys(self):
        get_match_keys = SearchNssetView.get_match_keys
        self.assertEqual(
            get_match_keys([]),
            {},
        )
        self.assertEqual(
            get_match_keys(['dns_hosts.fqdn', 'dns_hosts.ip_addresses']),
            {'dns_hosts': ['dns_hosts.fqdn', 'dns_hosts.ip_addresses']},
        )


@override_settings(ROOT_URLCONF='ferda.tests.urls')
class TestSearchKeysetView(LoggerMixin, TestCase):
    """Test SearchKeysetView."""

    def setUp(self):
        super().setUp()
        self.user = User.objects.create_superuser('admin', 'admin@example.com', 'password')
        self.client.force_login(self.user)

    @patch('ferda.backend.BACKEND.get_keyset_info')
    @patch('ferda.backend.BACKEND.search_keyset')
    def test_get(self, search_mock, keyset_mock):
        search_mock.return_value = {
            'results': [{
                'object_id': UUID(int=42),
                'matched_items': ['keyset_handle'],
            }],
            'searched_items': ['keyset_handle', 'dns_keys.key'],
        }
        keyset_mock.return_value = {
            'keyset_handle': 'KEYSET',
            'dns_keys': {
                'flags': 257,
                'protocol': 3,
                'alg': 5,
                'key': 'BlahBlahBlahQ==',
            },
        }
        response = self.client.get(
            reverse('registry:search:ajax_keyset'),
            data={'query': ['KEYSET'], 'type': 'keyset'},
        )
        data = response.json()
        self.assertEqual(
            data, {
                'search_results': [{
                    'object_id': str(UUID(int=42)),
                    'keyset_handle': 'KEYSET',
                    'dns_keys': {
                        'flags': 257,
                        'protocol': 3,
                        'alg': 5,
                        'key': 'BlahBlahBlahQ==',
                    },
                    'technical_contacts': [],
                    'matched_items': ['keyset_handle'],
                    'match_keys': {'keyset_handle': ['keyset_handle']},
                }]
            }
        )
        # Check logs
        properties = {'query_values': [LogProperty('KEYSET', None)], 'search_item': []}
        self.assertLog(
            LogEntryType.KEYSET_SEARCH,
            LoggerResult.SUCCESS,
            input_properties=properties,
            references={'keyset': [str(UUID(int=42))]},
        )

    def test_get_empty(self):
        response = self.client.get(reverse('registry:search:ajax_keyset'))
        self.assertJSONEqual(response.content, {'search_results': []})
        self.assertNoLogs()

    @override_settings(USE_TZ=True)
    @patch('ferda.backend.BACKEND.get_keyset_info', autospec=True)
    @patch('ferda.backend.BACKEND.get_keyset_history', autospec=True)
    @patch('ferda.backend.BACKEND.search_keyset_history', autospec=True)
    def test_get_history(self, search_mock, history_mock, info_mock):
        search_mock.return_value = {
            'results': [{
               'object_id': UUID(int=42),
               'histories': [{
                   'matched_items': ['keyset_handle'],
                   'valid_from': datetime(2019, 1, 1, tzinfo=pytz.utc),
                   'valid_to': datetime(2020, 1, 1, tzinfo=pytz.utc),
                   'keyset_history_ids': [UUID(int=142)],
               }],
            }],
        }
        history_mock.return_value = {
            'timeline': [{
                'keyset_history_id': UUID(int=142),
                'valid_from': datetime(2018, 1, 1, tzinfo=pytz.utc),
            }, {
                'keyset_history_id': UUID(int=42),
                'valid_from': datetime(2019, 1, 1, tzinfo=pytz.utc),
            }],
            'valid_to': None,
        }
        info_mock.return_value = {
            'keyset_handle': 'DUNE-KEYSET',
            'dns_keys': {
                'flags': 257,
                'protocol': 3,
                'alg': 5,
                'key': 'BlahBlahBlahQ==',
            },
            'technical_contacts': [],
            'events': {
                'registered': {
                    'registrar_handle': 'REG-DUNE',
                    'timestamp': datetime(2018, 5, 5, tzinfo=pytz.utc),
                },
                'unregistered': None,
            },
        }
        response = self.client.get(
            reverse('registry:search:ajax_keyset_history'),
            data={
                'query': 'dune',
                'type': 'keyset',
                'valid_from': '2019-01-01T00:00:00Z',
                'valid_to': '2020-01-01T00:00:00Z',
            },
        )
        data = response.json()
        self.assertEqual(data, {
            'search_results': [{
                'object_id': str(UUID(int=42)),
                'keyset_handle': 'DUNE-KEYSET',
                'technical_contacts': [],
                'dns_keys': {
                    'flags': 257,
                    'protocol': 3,
                    'alg': 5,
                    'key': 'BlahBlahBlahQ==',
                },
                'events': {
                    'registered': {
                        'registrar_handle': 'REG-DUNE',
                        'timestamp': '2018-05-05T00:00:00Z',
                    },
                    'unregistered': None,
                },
                'histories': [{
                    'keyset_handle': 'DUNE-KEYSET',
                    'keyset_history_ids': [str(UUID(int=142))],
                    'match_keys': {'keyset_handle': ['keyset_handle']},
                    'matched_items': ['keyset_handle'],
                    'valid_from': '2019-01-01T00:00:00Z',
                    'valid_to': '2020-01-01T00:00:00Z',
                    'technical_contacts': [],
                    'dns_keys': {
                        'flags': 257,
                        'protocol': 3,
                        'alg': 5,
                        'key': 'BlahBlahBlahQ==',
                    },
                }],
            }],
        })

        # Check logs
        properties = {'query_values': [LogProperty('dune', None)], 'search_item': []}
        self.assertLog(
            LogEntryType.KEYSET_SEARCH_HISTORY,
            LoggerResult.SUCCESS,
            input_properties=properties,
            references={'keyset': [str(UUID(int=42))]},
        )


class TestRelatedContactsMixin(LoggerMixin, TestCase):

    @patch('ferda.backend.BACKEND.get_contact_info_future')
    def test_get_technical_contacts_overview(self, info_mock):
        info_mock.return_value.result.return_value = {
            'contact_id': UUID(int=42),
            'contact_handle': 'IDAHO',
            'name': 'Duncan Idaho',
            'email': 'duncan.idaho@example.com',
            'events': {
                'registered': {
                    'registrar_handle': 'REG-DUNE',
                    'timestamp': datetime(2018, 5, 5, tzinfo=pytz.utc),
                },
                'unregistered': None,
            },
        }

        overview = RelatedContactsMixin.get_related_contacts_overview([UUID(int=42)])

        self.assertEqual(overview, [{'contact_handle': 'IDAHO', 'contact_id': UUID(int=42), 'name': 'Duncan Idaho'}])
